from emmet.core.utils import ValueEnum
from datetime import datetime
from monty.json import MontyDecoder


def convert_datetime(cls, v):
    if isinstance(v, dict):
        if v.get("$date"):
            return datetime.fromisoformat(v["$date"])

    return MontyDecoder().process_decoded(v)


class Status(ValueEnum):
    """
    State of a calculation/analysis.
    """

    SUCCESS = "successful"
    FAILED = "failed"



from pydantic import Field
from pydantic.main import BaseModel
from pymatgen.core.structure import Structure
from enum import Enum
from typing import Optional


class MPCompleteDoc(BaseModel):
    """
    Defines data for MPComplete structure submissions
    """

    structure: Optional[Structure] = Field(
        None,
        description="Structure submitted by the user.",
    )

    public_name: Optional[str] = Field(
        None,
        description="Public name of submitter.",
    )

    public_email: Optional[str] = Field(
        None,
        description="Public email of submitter.",
    )


class MPCompleteDataStatus(Enum):
    """
    Submission status for MPComplete data
    """

    submitted = "SUBMITTED"
    pending = "PENDING"
    running = "RUNNING"
    error = "ERROR"
    complete = "COMPLETE"



from typing import Dict, List, Optional

from pydantic import BaseModel, Field


class EOSDoc(BaseModel):
    """
    Fitted equations of state and energies and volumes used for fits.
    """

    energies: Optional[List[float]] = Field(
        None,
        description="Common energies in eV/atom that the equations of state are plotted with.",
    )

    volumes: Optional[List[float]] = Field(
        None,
        description="Common volumes in A³/atom that the equations of state are plotted with.",
    )

    eos: Optional[Dict] = Field(
        None,
        description="Data for each type of equation of state.",
    )

    material_id: Optional[str] = Field(
        None,
        description="The Materials Project ID of the material. This comes in the form: mp-******.",
    )



from pydantic import Field
from hashlib import blake2b
from typing import Optional

from emmet.core.mpid import MPculeID
from emmet.core.qchem.calc_types import LevelOfTheory
from emmet.core.qchem.task import TaskDocument
from emmet.core.material import PropertyOrigin
from emmet.core.molecules.molecule_property import PropertyDoc


__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


def get_free_energy(energy, enthalpy, entropy, temperature=298.15, convert_energy=True):
    """
    Helper function to calculate Gibbs free energy from electronic energy, enthalpy, and entropy

    :param energy: Electronic energy in Ha
    :param enthalpy: Enthalpy in kcal/mol
    :param entropy: Entropy in csal/mol-K
    :param temperature: Temperature in K. Default is 298.15, 25C

    returns: Free energy in eV

    """

    if convert_energy:
        e = energy * 27.2114
    else:
        e = energy

    return e + enthalpy * 0.043363 - temperature * entropy * 0.000043363


class MoleculeThermoDoc(PropertyDoc):
    property_name: str = "thermo"

    electronic_energy: float = Field(
        ..., description="Electronic energy of the molecule (units: eV)"
    )

    correction: bool = Field(
        False,
        description="Was a single-point calculation at higher level of "
        "theory used to correct the electronic energy?",
    )

    base_level_of_theory: Optional[LevelOfTheory] = Field(
        None, description="Level of theory used for uncorrected thermochemistry."
    )

    base_solvent: Optional[str] = Field(
        None,
        description="String representation of the solvent "
        "environment used for uncorrected thermochemistry.",
    )

    base_lot_solvent: Optional[str] = Field(
        None,
        description="String representation of the level of theory and solvent "
        "environment used for uncorrected thermochemistry.",
    )

    correction_level_of_theory: Optional[LevelOfTheory] = Field(
        None, description="Level of theory used to correct the electronic energy."
    )

    correction_solvent: Optional[str] = Field(
        None,
        description="String representation of the solvent "
        "environment used to correct the electronic energy.",
    )

    correction_lot_solvent: Optional[str] = Field(
        None,
        description="String representation of the level of theory and solvent "
        "environment used to correct the electronic energy.",
    )

    combined_lot_solvent: Optional[str] = Field(
        None,
        description="String representation of the level of theory and solvent "
        "environment used to generate this ThermoDoc, combining "
        "both the frequency calculation and (potentially) the "
        "single-point energy correction.",
    )

    zero_point_energy: Optional[float] = Field(
        None, description="Zero-point energy of the molecule (units: eV)"
    )

    rt: Optional[float] = Field(
        None,
        description="R*T, where R is the gas constant and T is temperature, taken "
        "to be 298.15K (units: eV)",
    )

    total_enthalpy: Optional[float] = Field(
        None, description="Total enthalpy of the molecule at 298.15K (units: eV)"
    )
    total_entropy: Optional[float] = Field(
        None, description="Total entropy of the molecule at 298.15K (units: eV/K)"
    )

    translational_enthalpy: Optional[float] = Field(
        None,
        description="Translational enthalpy of the molecule at 298.15K (units: eV)",
    )
    translational_entropy: Optional[float] = Field(
        None,
        description="Translational entropy of the molecule at 298.15K (units: eV/K)",
    )
    rotational_enthalpy: Optional[float] = Field(
        None, description="Rotational enthalpy of the molecule at 298.15K (units: eV)"
    )
    rotational_entropy: Optional[float] = Field(
        None, description="Rotational entropy of the molecule at 298.15K (units: eV/K)"
    )
    vibrational_enthalpy: Optional[float] = Field(
        None, description="Vibrational enthalpy of the molecule at 298.15K (units: eV)"
    )
    vibrational_entropy: Optional[float] = Field(
        None, description="Vibrational entropy of the molecule at 298.15K (units: eV/K)"
    )

    free_energy: Optional[float] = Field(
        None, description="Gibbs free energy of the molecule at 298.15K (units: eV)"
    )

    @classmethod
    def from_task(
        cls,
        task: TaskDocument,
        molecule_id: MPculeID,
        correction_task: Optional[TaskDocument] = None,
        deprecated: bool = False,
        **kwargs,
    ):  # type: ignore[override]
        """
        Construct a thermodynamics document from a task

        :param task: document from which thermodynamic properties can be extracted
        :param molecule_id: MPculeID
        :param deprecated: bool. Is this document deprecated?
        :param kwargs: to pass to PropertyDoc
        :return:
        """

        if task.output.optimized_molecule is not None:
            mol = task.output.optimized_molecule
        else:
            mol = task.output.initial_molecule

        if correction_task is None:
            energy = task.output.final_energy
            correction = False
            correction_lot = None
            correction_solvent = None
            correction_lot_solvent = None
            level_of_theory = task.level_of_theory
            solvent = task.solvent
            lot_solvent = task.lot_solvent
            combined_lot_solvent = task.lot_solvent
        else:
            energy = correction_task.output.final_energy
            correction = True
            correction_lot = correction_task.level_of_theory
            correction_solvent = correction_task.solvent
            correction_lot_solvent = correction_task.lot_solvent
            combined_lot_solvent = f"{task.lot_solvent}//{correction_lot_solvent}"
            level_of_theory = correction_lot
            solvent = correction_solvent
            lot_solvent = combined_lot_solvent

        total_enthalpy = task.output.enthalpy
        total_entropy = task.output.entropy

        origins = [PropertyOrigin(name="thermo", task_id=task.task_id)]
        id_string = f"thermo-{molecule_id}-{task.task_id}-{task.lot_solvent}"
        if correction and correction_task is not None:
            origins.append(
                PropertyOrigin(
                    name="thermo_energy_correction", task_id=correction_task.task_id
                )
            )

            id_string += f"-{correction_task.task_id}-{correction_task.lot_solvent}"

        h = blake2b()
        h.update(id_string.encode("utf-8"))
        property_id = h.hexdigest()

        if total_enthalpy is not None and total_entropy is not None:
            free_energy = get_free_energy(energy, total_enthalpy, total_entropy)

            for calc in task.calcs_reversed:
                if all(
                    [
                        calc.get(x) is not None
                        for x in [
                            "ZPE",
                            "trans_enthalpy",
                            "rot_enthalpy",
                            "vib_enthalpy",
                            "gas_constant",
                            "trans_entropy",
                            "rot_entropy",
                            "vib_entropy",
                        ]
                    ]
                ):
                    return super().from_molecule(
                        meta_molecule=mol,
                        property_id=property_id,
                        molecule_id=molecule_id,
                        level_of_theory=level_of_theory,
                        solvent=solvent,
                        lot_solvent=lot_solvent,
                        correction=correction,
                        base_level_of_theory=task.level_of_theory,
                        base_solvent=task.solvent,
                        base_lot_solvent=task.lot_solvent,
                        correction_level_of_theory=correction_lot,
                        correction_solvent=correction_solvent,
                        correction_lot_solvent=correction_lot_solvent,
                        combined_lot_solvent=combined_lot_solvent,
                        electronic_energy=energy * 27.2114,
                        zero_point_energy=calc["ZPE"] * 0.043363,
                        rt=calc["gas_constant"] * 0.043363,
                        total_enthalpy=total_enthalpy * 0.043363,
                        translational_enthalpy=calc["trans_enthalpy"] * 0.043363,
                        rotational_enthalpy=calc["rot_enthalpy"] * 0.043363,
                        vibrational_enthalpy=calc["vib_enthalpy"] * 0.043363,
                        total_entropy=total_entropy * 0.000043363,
                        translational_entropy=calc["trans_entropy"] * 0.000043363,
                        rotational_entropy=calc["rot_entropy"] * 0.000043363,
                        vibrational_entropy=calc["vib_entropy"] * 0.000043363,
                        free_energy=free_energy,
                        deprecated=deprecated,
                        origins=origins,
                        **kwargs,
                    )

        # If all thermodynamic data is not available
        return super().from_molecule(
            meta_molecule=mol,
            property_id=property_id,
            molecule_id=molecule_id,
            level_of_theory=task.level_of_theory,
            solvent=task.solvent,
            lot_solvent=task.lot_solvent,
            correction=correction,
            correction_level_of_theory=correction_lot,
            correction_solvent=correction_solvent,
            correction_lot_solvent=correction_lot_solvent,
            combined_lot_solvent=combined_lot_solvent,
            electronic_energy=energy * 27.2114,
            deprecated=deprecated,
            origins=origins,
            **kwargs,
        )



from pydantic import BaseModel, Field
from typing import Optional


class SubstratesDoc(BaseModel):
    """
    Possible growth substrates for a given material.
    """

    sub_form: Optional[str] = Field(
        None,
        description="Reduced formula of the substrate.",
    )

    sub_id: Optional[str] = Field(
        None,
        description="Materials Project ID of the substrate material. This comes in the form: mp-******.",
    )

    film_orient: Optional[str] = Field(
        None,
        description="Surface orientation of the film material.",
    )

    area: Optional[float] = Field(
        None,
        description="Minimum coincident interface area in Å².",
    )

    energy: Optional[float] = Field(
        None,
        description="Elastic energy in meV.",
    )

    film_id: Optional[str] = Field(
        None,
        description="The Materials Project ID of the film material. This comes in the form: mp-******.",
    )

    norients: Optional[int] = Field(
        None,
        description="Number of possible surface orientations for the substrate.",
        alias="_norients",
    )

    orient: Optional[str] = Field(
        None,
        description="Surface orientation of the substrate material.",
    )



from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel, Field, field_validator

from emmet.core.common import convert_datetime


class FermiDoc(BaseModel):
    """
    Fermi surfaces.
    """

    fermi_surfaces: Optional[List[dict]] = Field(
        None,
        description="List of IFermi FermiSurface objects.",
    )

    surface_types: Optional[List[str]] = Field(
        None,
        description="Type of each fermi surface in the fermi_surfaces list.\
            Is either CBM or VBM for semiconductors, or fermi_surface for metals.",
    )

    material_id: Optional[str] = Field(
        None,
        description="The Materials Project ID of the material. This comes in the form: mp-******.",
    )

    last_updated: Optional[datetime] = Field(
        None,
        description="Timestamp for the most recent calculation for this fermi surface document.",
    )

    # Make sure that the datetime field is properly formatted
    @field_validator("last_updated", mode="before")
    @classmethod
    def handle_datetime(cls, v):
        return convert_datetime(cls, v)



# mypy: ignore-errors

"""
Settings for defaults in the core definitions of Materials Project Documents
"""
import json
from pathlib import Path
from typing import Dict, List, Type, TypeVar, Union

import requests
from monty.json import MontyDecoder
from pydantic import field_validator, model_validator, Field, ImportString
from pydantic_settings import BaseSettings, SettingsConfigDict

DEFAULT_CONFIG_FILE_PATH = str(Path.home().joinpath(".emmet.json"))


S = TypeVar("S", bound="EmmetSettings")


class EmmetSettings(BaseSettings):
    """
    Settings for the emmet- packages
    Non-core packages should subclass this to get settings specific to their needs
    The default way to modify these is to modify ~/.emmet.json or set the environment variable
    EMMET_CONFIG_FILE to point to the json with emmet settings
    """

    config_file: str = Field(
        DEFAULT_CONFIG_FILE_PATH, description="File to load alternative defaults from"
    )

    LTOL: float = Field(
        0.2, description="Fractional length tolerance for structure matching"
    )
    STOL: float = Field(
        0.3,
        description="Site tolerance for structure matching. Defined as the fraction of the"
        " average free length per atom = ( V / Nsites ) ** (1/3)",
    )
    SYMPREC: float = Field(
        0.1, description="Symmetry precision for spglib symmetry finding"
    )
    ANGLE_TOL: float = Field(
        5, description="Angle tolerance for structure matching in degrees."
    )

    PGATOL: float = Field(
        0.3,
        description="Distance tolerance to consider sites as symmetrically equivalent.",
    )
    PGAEIGENTOL: float = Field(
        0.01, description="Tolerance to compare eigen values of the inertia tensor."
    )
    PGAMATRIXTOL: float = Field(
        0.1,
        description="Tolerance used to generate the full set of symmetry operations of the point group.",
    )

    MAX_PIEZO_MILLER: int = Field(
        10,
        description="Maximum miller allowed for computing strain direction for maximal piezo response",
    )

    QCHEM_FUNCTIONAL_QUALITY_SCORES: Dict[str, int] = Field(
        {
            "wB97M-V": 7,
            "wB97X-V": 6,
            "wB97X-D3": 5,
            "wB97X-D": 5,
            "B3LYP": 4,
            "B97M-V": 3,
            "B97M-rV": 3,
            "B97-D3": 2,
            "B97-D": 2,
            "PBE": 1,
        },
        description="Dictionary mapping Q-Chem density functionals to a quality score.",
    )

    QCHEM_BASIS_QUALITY_SCORES: Dict[str, int] = Field(
        {
            "6-31g*": 1,
            "def2-SVPD": 2,
            "def2-TZVP": 3,
            "def2-TZVPD": 4,
            "def2-TZVPP": 5,
            "def2-TZVPPD": 6,
            "def2-QZVPPD": 7,
        },
        description="Dictionary mapping Q-Chem basis sets to a quality score.",
    )

    QCHEM_SOLVENT_MODEL_QUALITY_SCORES: Dict[str, int] = Field(
        {"CMIRS": 7, "SMD": 5, "ISOSVP": 4, "PCM": 3, "VACUUM": 1},
        description="Dictionary mapping Q-Chem solvent models to a quality score.",
    )

    QCHEM_TASK_QUALITY_SCORES: Dict[str, int] = Field(
        {
            "single_point": 1,
            "geometry optimization": 2,
            "frequency-flattening geometry optimization": 3,
        },
        description="Dictionary mapping Q-Chem task type to a quality score",
    )

    VASP_STRUCTURE_QUALITY_SCORES: Dict[str, int] = Field(
        {"R2SCAN": 5, "SCAN": 4, "GGA+U": 3, "GGA": 2, "PBESol": 1},
        description="Dictionary Mapping VASP calculation run types to rung level for VASP materials builder structure data",  # noqa: E501
    )

    VASP_KPTS_TOLERANCE: float = Field(
        0.9,
        description="Relative tolerance for kpt density to still be a valid task document",
    )

    VASP_KSPACING_TOLERANCE: float = Field(
        0.05,
        description="Relative tolerance for kspacing to still be a valid task document",
    )

    VASP_DEFAULT_INPUT_SETS: Dict[str, ImportString] = Field(
        {
            "GGA Structure Optimization": "pymatgen.io.vasp.sets.MPRelaxSet",
            "GGA+U Structure Optimization": "pymatgen.io.vasp.sets.MPRelaxSet",
            "R2SCAN Structure Optimization": "pymatgen.io.vasp.sets.MPScanRelaxSet",
            "SCAN Structure Optimization": "pymatgen.io.vasp.sets.MPScanRelaxSet",
            "PBESol Structure Optimization": "pymatgen.io.vasp.sets.MPScanRelaxSet",
            "GGA Static": "pymatgen.io.vasp.sets.MPStaticSet",
            "GGA+U Static": "pymatgen.io.vasp.sets.MPStaticSet",
            "R2SCAN Static": "pymatgen.io.vasp.sets.MPScanStaticSet",
            "SCAN Static": "pymatgen.io.vasp.sets.MPScanStaticSet",
            "PBESol Static": "pymatgen.io.vasp.sets.MPScanStaticSet",
            "HSE06 Static": "pymatgen.io.vasp.sets.MPScanStaticSet",
            "GGA NSCF Uniform": "pymatgen.io.vasp.sets.MPNonSCFSet",
            "GGA+U NSCF Uniform": "pymatgen.io.vasp.sets.MPNonSCFSet",
            "GGA NSCF Line": "pymatgen.io.vasp.sets.MPNonSCFSet",
            "GGA+U NSCF Line": "pymatgen.io.vasp.sets.MPNonSCFSet",
            "GGA NMR Electric Field Gradient": "pymatgen.io.vasp.sets.MPNMRSet",
            "GGA NMR Nuclear Shielding": "pymatgen.io.vasp.sets.MPNMRSet",
            "GGA+U NMR Electric Field Gradient": "pymatgen.io.vasp.sets.MPNMRSet",
            "GGA+U NMR Nuclear Shielding": "pymatgen.io.vasp.sets.MPNMRSet",
            "GGA Deformation": "pymatgen.io.vasp.sets.MPStaticSet",
            "GGA+U Deformation": "pymatgen.io.vasp.sets.MPStaticSet",
            "GGA DFPT Dielectric": "pymatgen.io.vasp.sets.MPStaticSet",
            "GGA+U DFPT Dielectric": "pymatgen.io.vasp.sets.MPStaticSet",
        },
        description="Default input sets for task validation",
    )

    VASP_VALIDATE_POTCAR_STATS: bool = Field(
        True, description="Whether to validate POTCAR stat values."
    )

    VASP_CHECKED_LDAU_FIELDS: List[str] = Field(
        ["LDAUU", "LDAUJ", "LDAUL"], description="LDAU fields to validate for tasks"
    )

    VASP_MAX_SCF_GRADIENT: float = Field(
        100,
        description="Maximum upward gradient in the last SCF for any VASP calculation",
    )

    VASP_USE_STATICS: bool = Field(
        True,
        description="Use static calculations for structure and energy along with structure optimizations",
    )
    model_config = SettingsConfigDict(env_prefix="emmet_", extra="ignore")

    @model_validator(mode="before")
    @classmethod
    def load_default_settings(cls, values):
        """
        Loads settings from a root file if available and uses that as defaults in
        place of built in defaults
        """
        config_file_path: str = values.get("config_file", DEFAULT_CONFIG_FILE_PATH)

        new_values = {}

        if config_file_path.startswith("http"):
            new_values = requests.get(config_file_path).json()
        elif Path(config_file_path).exists():
            with open(config_file_path) as f:
                new_values = json.load(f)

        new_values.update(values)

        return new_values

    @classmethod
    def autoload(cls: Type[S], settings: Union[None, dict, S]) -> S:
        if settings is None:
            return cls()
        elif isinstance(settings, dict):
            return cls(**settings)
        return settings

    @field_validator("VASP_DEFAULT_INPUT_SETS", mode="before")
    @classmethod
    def convert_input_sets(cls, value):
        if isinstance(value, dict):
            return {k: MontyDecoder().process_decoded(v) for k, v in value.items()}
        return value

    def as_dict(self):
        """
        HotPatch to enable serializing EmmetSettings via Monty
        """
        return self.dict(exclude_unset=True, exclude_defaults=True)



from pathlib import Path

try:
    import emmet.core.qchem.calc_types.enums
except ImportError:
    import emmet.core.qchem.calc_types.generate

from emmet.core.qchem.calc_types.enums import CalcType, LevelOfTheory, TaskType
from emmet.core.qchem.calc_types.utils import (
    calc_type,
    level_of_theory,
    task_type,
    solvent,
    lot_solvent_string,
)



# %%
import re
from typing import Union, Any, Callable

from pydantic_core import CoreSchema, core_schema

from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue


# matches "mp-1234" or "1234" followed by and optional "-(Alphanumeric)"
mpid_regex = re.compile(r"^([A-Za-z]*-)?(\d+)(-[A-Za-z0-9]+)*$")
mpculeid_regex = re.compile(
    r"^([A-Za-z]+-)?([A-Fa-f0-9]+)-([A-Za-z0-9]+)-(m?[0-9]+)-([0-9]+)$"
)
# matches capital letters and numbers of length 26 (ULID)
# followed by and optional "-(Alphanumeric)"
check_ulid = re.compile(r"^[A-Z0-9]{26}(-[A-Za-z0-9]+)*$")


class MPID(str):
    """
    A Materials Project type ID with a prefix and an integer
    This class enables seemlessly mixing MPIDs and regular integer IDs
    Prefixed IDs are considered less than non-prefixed IDs to enable proper
    mixing with the Materials Project

    Args:
        val: Either 1) a prefixed string e.g. "mp-1234"
                    2) an integer e.g. 1234
                    3) a number stored as a string e.g. '1234'
                    4) an MPID

            Numbers stored as strings are coerced to ints
    """

    def __init__(self, val: Union["MPID", int, str]):
        if isinstance(val, MPID):
            self.parts = val.parts  # type: ignore
            self.string = val.string  # type: ignore

        elif isinstance(val, int) or (isinstance(val, str) and val.isnumeric()):
            self.parts = ("", int(val))
            self.string = str(val)

        elif isinstance(val, str):
            if mpid_regex.fullmatch(val):
                parts = val.split("-")
                parts[1] = int(parts[1])  # type: ignore
                self.parts = tuple(parts)
            elif check_ulid.fullmatch(val):
                ulid = val.split("-")[0]
                self.parts = (ulid, 0)
            else:
                raise ValueError(
                    "MPID string representation must follow the format prefix-number or start with a valid ULID."
                )
            self.string = val

        else:
            raise ValueError(
                "Must provide an MPID, int, or string of the format prefix-number or start with a valid ULID."
            )

    def __eq__(self, other: object):
        if isinstance(other, MPID):
            return self.string == other.string
        elif isinstance(other, (int, str)):
            return self.string == MPID(other).string

    def __str__(self):
        return self.string

    def __repr__(self):
        return f"MPID({self})"

    def __lt__(self, other: Union["MPID", int, str]):
        other_parts = MPID(other).parts

        if self.parts[0] != "" and other_parts[0] != "":
            # both have prefixes; normal comparison
            return self.parts < other_parts
        elif self.parts[0] != "":
            # other is a pure int, self is prefixed
            # Always sort MPIDs before pure integer IDs
            return True
        elif other_parts[0] != "":
            # self is pure int, other is prefixed
            return False
        else:
            # both are pure ints; normal comparison
            return self.parts[1] < other_parts[1]

    def __gt__(self, other: Union["MPID", int, str]):
        return not self.__lt__(other)

    def __hash__(self):
        return hash(self.string)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: Callable[[Any], core_schema.CoreSchema]
    ) -> core_schema.CoreSchema:
        return core_schema.with_info_plain_validator_function(cls.validate)

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return dict(
            pattern=r"^([A-Za-z]*-)?(\d+)(-[A-Za-z0-9]+)*$",
            examples=["mp-3534", "3453", "mp-834-Ag"],
            type="string",
        )

    @classmethod
    def validate(cls, __input_value: Any, _: core_schema.ValidationInfo):
        if isinstance(__input_value, MPID):
            return __input_value
        elif isinstance(__input_value, str) and mpid_regex.fullmatch(__input_value):
            return MPID(__input_value)
        elif isinstance(__input_value, str) and check_ulid.fullmatch(__input_value):
            return MPID(__input_value)
        elif isinstance(__input_value, int):
            return MPID(__input_value)

        raise ValueError("Invalid MPID Format")


# %%


class MPculeID(str):
    """
    A Materials Project Molecule ID with a prefix, hash, and two integer values
        representing the charge and spin of the molecule
    Unlike the MPID, you cannot compare raw hashes or raw integers to MPculeIDs

    Args:
        val: Either 1) an MPculeID
                    2) a prefixed string of format "prefix-hash-formula-charge-spin"
                    3) a non-prefixed string of format "hash-formula-charge-spin"

            Numbers stored as strings are coerced to ints
    """

    def __init__(self, val: Union["MPculeID", str]):
        if isinstance(val, MPculeID):
            self.parts = val.parts  # type: ignore
            self.string = val.string  # type: ignore

        elif isinstance(val, str):
            parts = val.split("-")
            if len(parts) == 4:
                parts[1] = int(parts[2].replace("m", "-"))  # type: ignore
                parts[2] = int(parts[3])  # type: ignore
            elif len(parts) == 5:
                parts[3] = int(parts[3].replace("m", "-"))  # type: ignore
                parts[4] = int(parts[4])  # type: ignore
            else:
                raise ValueError(
                    "MPculeID string representation must follow the "
                    "format prefix-hash-formula-charge-spin or hash-formula-charge-spin."
                )

            self.parts = tuple(parts)
            self.string = val

        else:
            raise ValueError(
                "Must provide an MPculeID, or string of the format prefix-hash-formula-charge-spin "
                "or hash-formula-charge-spin"
            )

    def __eq__(self, other: object):
        if isinstance(other, MPculeID):
            return self.string == other.string
        elif isinstance(other, str):
            return self.string == MPculeID(other).string

    def __str__(self):
        return self.string

    def __repr__(self):
        return f"MPculeID({self})"

    def __lt__(self, other: Union["MPculeID", str]):
        other_parts = MPculeID(other).parts

        return "-".join([str(x) for x in self.parts[-4:]]) < "-".join(
            [str(x) for x in other_parts[-4:]]
        )

    def __gt__(self, other: Union["MPculeID", str]):
        return not self.__lt__(other)

    def __hash__(self):
        return hash(self.string)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: Callable[[Any], core_schema.CoreSchema]
    ) -> core_schema.CoreSchema:
        return core_schema.with_info_plain_validator_function(cls.validate)

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return dict(
            pattern=r"^^([A-Za-z]+-)?([A-Fa-f0-9]+)-([A-Za-z0-9]+)-(m?[0-9]+)-([0-9]+)$",
            examples=[
                "1a525231bdac3f13e2fac0962fe8d053-Mg1-0-1",
                "22b40b99719ac570fc7e6225e855ec6e-F5Li1P1-m1-2",
                "mpcule-b9ba54febc77d2a9177accf4605767db-C1H41-2",
            ],
            type="string",
        )

    @classmethod
    def validate(cls, __input_value: Any, _: core_schema.ValidationInfo):
        if isinstance(__input_value, MPculeID):
            return __input_value
        elif isinstance(__input_value, str) and mpculeid_regex.fullmatch(__input_value):
            return MPculeID(__input_value)

        raise ValueError("Invalid MPculeID Format")
